from enum import Enum
from itertools import chain
from operator import attrgetter
from typing import List, Type

from .app_settings import app_settings
from .utils import get_all_subclasses


class SubscriptionBase(Enum):
    @classmethod
    def get_all_subscriptions(cls) -> List[Type["SubscriptionBase"]]:
        subscriptions = get_all_subclasses(cls)

        if len(subscriptions) != len(set(subscription for subscription in subscriptions)):
            raise ValueError("Subscriptions should have unique values.")

        return list(subscriptions)

    @classmethod
    def validate_chosen_subscriptions(cls):
        selected_subscriptions = app_settings.SUBSCRIPTIONS
        all_subscriptions = list(map(attrgetter("value"), chain.from_iterable(cls.get_all_subscriptions())))
        invalid_subscriptions = list(
            filter(
                lambda subscription: subscription not in all_subscriptions,
                selected_subscriptions,
            )
        )
        if invalid_subscriptions and selected_subscriptions:
            raise ValueError(f"Invalid subscriptions: {', '.join(map(str, invalid_subscriptions))}")



import contextlib
import json
import logging

try:
    from google.api_core.exceptions import NotFound
    from google.cloud import scheduler_v1
    from google.cloud.scheduler_v1.types.cloudscheduler import (
        CreateJobRequest,
        DeleteJobRequest,
        GetJobRequest,
        ListJobsRequest,
        ListJobsResponse,
        UpdateJobRequest,
    )
    from google.cloud.scheduler_v1.types.job import Job
    from google.cloud.scheduler_v1.types.target import PubsubTarget
except ImportError:
    scheduler_v1 = None

from .app_settings import app_settings
from .types import SchedulerJob

logger = logging.getLogger("flex_pubsub")


class BaseSchedulerBackend:
    def delete_job(self, task_name: str) -> None:
        raise NotImplementedError

    def list_jobs(self) -> ListJobsResponse:
        raise NotImplementedError

    def schedule(self, task_name: str, schedule_config: SchedulerJob) -> Job:
        raise NotImplementedError


class LocalSchedulerBackend(BaseSchedulerBackend):
    def delete_job(self, task_name: str) -> None:
        logger.info(
            f"LocalSchedulerBackend: delete_job called with task_name={task_name}"
        )

    def list_jobs(self) -> ListJobsResponse:
        logger.info("LocalSchedulerBackend: list_jobs called")
        return ListJobsResponse()

    def schedule(self, task_name: str, schedule_config: SchedulerJob) -> Job:
        logger.info(
            f"LocalSchedulerBackend: schedule called with task_name={task_name}, schedule_config={schedule_config}"
        )
        return Job(name=task_name)


class GoogleSchedulerBackend(BaseSchedulerBackend):
    def __init__(self) -> None:
        if scheduler_v1 is None:
            raise ImportError("google-cloud-scheduler is not installed.")
        credentials = app_settings.GOOGLE_CREDENTIALS
        self.client = scheduler_v1.CloudSchedulerClient(credentials=credentials)
        self.project_id = app_settings.GOOGLE_PROJECT_ID
        self.location = app_settings.SCHEDULER_LOCATION

        self.parent = self._get_parent()
        logger.info("Initialized GoogleSchedulerBackend")

    def delete_job(self, task_name: str) -> None:
        job_name = self._get_job_name(task_name)
        with contextlib.suppress(NotFound):
            self.client.delete_job(request=DeleteJobRequest(name=job_name))

    def list_jobs(self) -> ListJobsResponse:
        return self.client.list_jobs(request=ListJobsRequest(parent=self.parent))

    def schedule(self, task_name: str, schedule_config: SchedulerJob) -> Job:
        job = Job(
            name=self._get_job_name(task_name),
            schedule=schedule_config.schedule,
            time_zone=schedule_config.time_zone,
            pubsub_target=PubsubTarget(
                topic_name=app_settings.TOPIC_PATH,
                data=json.dumps(
                    {
                        "task_name": task_name,
                        "args": schedule_config.args,
                        "kwargs": schedule_config.kwargs,
                    }
                ).encode(),
            ),
        )
        return self._get_or_create_or_update_task(job)

    def _get_job_name(self, task_name: str) -> str:
        return f"{self.parent}/jobs/{task_name}"

    def _get_parent(self):
        return f"projects/{self.project_id}/locations/{self.location}"

    def _get_job(self, request: GetJobRequest) -> Job | None:
        with contextlib.suppress(NotFound):
            return self.client.get_job(request=request)

    def _compare_jobs(self, retrieved_job: Job, job: Job) -> bool:
        return (
            retrieved_job.schedule == job.schedule
            and retrieved_job.time_zone == job.time_zone
            and retrieved_job.pubsub_target == job.pubsub_target
        )

    def _get_or_create_or_update_task(self, job: Job) -> Job:
        if retrieved_job := self._get_job(GetJobRequest(name=job.name)):
            if not self._compare_jobs(retrieved_job, job):
                return self.client.update_job(request=UpdateJobRequest(job=job))
            return retrieved_job

        return self.client.create_job(
            request=CreateJobRequest(parent=self.parent, job=job)
        )



from typing import Any, Dict, List, Optional

from .app_settings import app_settings
from .types import RequestMessage


def send_task(
    *args: Optional[List[Any]],
    task_name: str,
    **kwargs: Optional[Dict[str, Any]],
) -> None:
    args = args or []
    kwargs = kwargs or {}

    raw_message = {
        "task_name": task_name,
        "args": args,
        "kwargs": kwargs,
    }
    message_str = RequestMessage.model_validate(raw_message).model_dump_json()
    backend_class = app_settings.BACKEND_CLASS
    backend = backend_class()
    backend.publish(message_str)



from django.apps import AppConfig


class FlexPubsubConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "flex_pubsub"

    def ready(self):
        from . import checks  # noqa: F401



from enum import StrEnum
from typing import List

from .app_settings import app_settings


def are_subscriptions_valid(subscriptions: List[StrEnum]):
    if app_settings.SUBSCRIPTIONS:
        return any(i.value in app_settings.SUBSCRIPTIONS for i in subscriptions)
    return True


def get_all_subclasses(klass):
    subclasses = set()
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses



from typing import Optional

from pydantic import BaseModel, Field

from ..app_settings import app_settings


class SchedulerJob(BaseModel):
    schedule: str
    time_zone: Optional[str] = Field(default="PST")
    pubsub_topic: str = Field(default=app_settings.TOPIC_NAME)
    args: Optional[list] = Field(default=[])
    kwargs: Optional[dict] = Field(default={})



from django.core.checks import register

from .subscription import SubscriptionBase
from .tasks import task_registry


@register()
def sync_scheduler_configs(app_configs, **kwargs):
    task_registry.sync_registered_jobs()
    SubscriptionBase.validate_chosen_subscriptions()
    return []



from logging import getLogger
from typing import Any

from django.core.management.base import BaseCommand

from flex_pubsub.app_settings import app_settings
from flex_pubsub.backends import BaseBackend
from flex_pubsub.tasks import task_registry
from flex_pubsub.types import CallbackContext, RequestMessage

logger = getLogger("django.pubsub")


class Command(BaseCommand):
    help = "Starts the subscriber to listen for messages and execute tasks."

    def message_callback(self, context: CallbackContext) -> None:
        raw_message = context.raw_message
        logger.info(f"Received message: {raw_message}")

        ack = context.ack
        data = RequestMessage.model_validate_json(raw_message)

        task = task_registry.get_task(data.task_name)
        t_args = data.args
        t_kwargs = data.kwargs

        if not task or context.subscription_name not in task.subscriptions:
            ack()
            return

        if set(task.subscriptions).issubset(set(app_settings.SUBSCRIPTIONS)):
            try:
                task(*t_args, **t_kwargs)
            except Exception as e:
                logger.error(
                    f"Error executing task: {e}",
                    exc_info=True,
                    extra={
                        "task_name": data.task_name,
                        "args": t_args,
                        "kwargs": t_kwargs,
                    },
                )

            ack()

    def display_registered_tasks(self):
        self.stdout.write("Registered tasks:")
        for task_name in task_registry.get_all_tasks():
            self.stdout.write(
                f"  - {task_name} ({schedule if (schedule:=task_registry.get_schedule_config(task_name)) else 'No schedule'})"
            )

    def handle(self, *args: Any, **options: Any) -> None:
        backend_class = app_settings.BACKEND_CLASS
        backend: BaseBackend = backend_class()

        self.display_registered_tasks()
        self.stdout.write("Starting subscriber...")
        backend.subscribe(self.message_callback)



import http
import http.server
import logging
from operator import call
from typing import Any, Callable, Dict, Type

try:
    from google.api_core.exceptions import NotFound
    from google.cloud import pubsub_v1
except ImportError:
    pubsub_v1 = None

import threading

from .app_settings import app_settings
from .tasks import task_registry
from .types import CallbackContext, RequestMessage, SubscriptionCallback

logger = logging.getLogger("flex_pubsub")


class Singleton(type):
    _instances: Dict[Type, Any] = {}

    def __call__(cls, *args, **kwargs) -> Any:
        if cls not in cls._instances or kwargs.get("force_new_instance", False):
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class BaseBackend(metaclass=Singleton):
    def publish(self, message: RequestMessage) -> None:
        raise NotImplementedError

    def subscribe(self, callback: Callable[[CallbackContext], None]) -> None:
        raise NotImplementedError


class LocalPubSubBackend(BaseBackend):
    def publish(self, raw_message: RequestMessage):
        logger.info(f"Publishing message: {raw_message}")

        message = RequestMessage.model_validate_json(raw_message)
        if not (task := task_registry.get_task(task_name := message.task_name)):
            raise ValueError(f"Task '{task_name}' not found.")

        threaded_call = threading.Thread(
            target=lambda: call(task, *message.args, **message.kwargs)
        )
        threaded_call.start()

    def subscribe(self, *args, **kwargs) -> None:
        logger.info("Subscribing to local pub/sub (Doing nothing)")


class GooglePubSubBackend(BaseBackend):
    def __init__(self) -> None:
        if pubsub_v1 is None:
            raise ImportError("google-cloud-pubsub is not installed.")
        credentials = app_settings.GOOGLE_CREDENTIALS
        project_id = app_settings.GOOGLE_PROJECT_ID

        self.subscriber = pubsub_v1.SubscriberClient(credentials=credentials)
        self.publisher = pubsub_v1.PublisherClient(credentials=credentials)
        self.subscriptions = {
            subscription_name: self.subscriber.subscription_path(
                project_id,
                subscription_name,
            )
            for subscription_name in app_settings.SUBSCRIPTIONS
        }
        self.topic_path = self.publisher.topic_path(project_id, app_settings.TOPIC_NAME)
        logger.info("Initialized GooglePubSubBackend")

    def run_server(
        self,
        server_class=http.server.HTTPServer,
        handler_class=http.server.BaseHTTPRequestHandler,
        port=app_settings.LISTENER_PORT,
    ):
        server_address = ("", port)
        httpd = server_class(server_address, handler_class)
        logger.info(f"Starting server on port {port}")
        httpd.serve_forever()

    def publish(self, message: RequestMessage) -> None:
        logger.info(f"Publishing message: {message}")

        request_message = RequestMessage.model_validate_json(message)
        self._ensure_topic_exists()

        logger.info(f"Publishing message to topic {self.topic_path}")
        self.publisher.publish(
            self.topic_path, request_message.model_dump_json().encode("utf-8")
        )

    def subscribe(self, callback: SubscriptionCallback) -> None:
        for subscription_name, subscription_path in self.subscriptions.items():
            self._ensure_subscription_exists(subscription_path)
            logger.info(f"Subscribing to {subscription_path}")
            self.subscriber.subscribe(
                subscription_path,
                callback=self._wrap_callback(callback, subscription_name),
            )

        self.run_server()

    def _wrap_callback(
        self,
        callback: SubscriptionCallback,
        subscription_name: str,
    ) -> Callable[[str], None]:
        from google.cloud.pubsub_v1.subscriber.message import Message

        def _callback(message: Message) -> None:
            context = CallbackContext(
                raw_message=message.data.decode("utf-8"),
                ack=message.ack,
                subscription_name=subscription_name,
            )
            callback(context)

        return _callback

    def _ensure_topic_exists(self) -> None:
        try:
            self.publisher.get_topic(request={"topic": self.topic_path})
        except NotFound:
            logger.warning(f"Topic not found: {self.topic_path}. Creating topic.")
            self.publisher.create_topic(request={"name": self.topic_path})

    def _ensure_subscription_exists(self, subscription_path: str) -> None:
        try:
            self.subscriber.get_subscription(
                request={"subscription": subscription_path}
            )
        except NotFound:
            logger.warning(
                f"Subscription not found: {subscription_path}. Creating subscription."
            )
            self.subscriber.create_subscription(
                request={"name": subscription_path, "topic": self.topic_path}
            )



import importlib

from django.conf import settings


def import_attribute(path):
    assert isinstance(path, str)
    module_path, attr_name = path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, attr_name)


def get_setting(name, default):
    return settings.PUBSUB_SETTINGS.get(name, default)


class AppSettings:
    def _setting(self, name, default):
        return get_setting(name, default) or default

    @property
    def BACKEND_CLASS(self):
        backend_class_path = self._setting("BACKEND_CLASS", "flex_pubsub.backends.LocalPubSubBackend")
        return import_attribute(backend_class_path)

    @property
    def SCHEDULER_BACKEND_CLASS(self):
        scheduler_backend_class_path = self._setting(
            "SCHEDULER_BACKEND_CLASS",
            "flex_pubsub.scheduler.LocalSchedulerBackend",
        )
        return import_attribute(scheduler_backend_class_path)

    @property
    def GOOGLE_CREDENTIALS(self):
        credentials = self._setting("GOOGLE_CREDENTIALS", None)
        if credentials:
            if isinstance(credentials, str):
                from google.oauth2 import service_account

                return service_account.Credentials.from_service_account_file(credentials)
            return credentials
        return None

    @property
    def GOOGLE_PROJECT_ID(self):
        return self._setting("GOOGLE_PROJECT_ID", None)

    @property
    def TOPIC_NAME(self):
        return self._setting("TOPIC_NAME", "default-topic")

    @property
    def TOPIC_PATH(self):
        return f"projects/{self.GOOGLE_PROJECT_ID}/topics/{self.TOPIC_NAME}"

    @property
    def SCHEDULER_LOCATION(self):
        return self._setting("SCHEDULER_LOCATION", "us-central1")

    @property
    def CLOUD_RUN_SERVICE_URL(self):
        return self._setting("CLOUD_RUN_SERVICE_URL", None)

    @property
    def LISTENER_PORT(self):
        return int(port) if str(port := self._setting("LISTENER_PORT", None)).isdigit() else 8001

    @property
    def SUBSCRIPTIONS(self):
        return list(map(str.strip, filter(bool, self._setting("SUBSCRIPTIONS", "").strip().split(","))))


app_settings = AppSettings()



from functools import partial, wraps
from typing import Any, Callable, Dict, List, Optional

from google.cloud.scheduler_v1.types.job import Job

from .app_settings import app_settings
from .constants import TASK_EXTRA_CONTEXT_ATTRIBUTE
from .scheduler import BaseSchedulerBackend
from .types import SchedulerJob
from .utils import are_subscriptions_valid


class TaskRegistry:
    def __init__(self) -> None:
        self.tasks: Dict[str, Callable] = {}
        self.schedule_configs: Dict[str, Job] = {}

    def register(
        self,
        func: Optional[Callable] = None,
        *,
        name: Optional[str] = None,
        raw_schedule: Optional[SchedulerJob] = None,
    ) -> Callable:
        if func is None:
            return lambda f: self.register(f, name=name, raw_schedule=raw_schedule)
        task_name = name or func.__name__
        self.tasks[task_name] = func
        if raw_schedule and (schedule := SchedulerJob.model_validate(raw_schedule)):
            self.schedule_configs[task_name] = schedule.model_dump()
            scheduler_backend: BaseSchedulerBackend = (
                app_settings.SCHEDULER_BACKEND_CLASS()
            )
            scheduler_backend.schedule(task_name, schedule)
        return func

    def _get_job_name(self, job: Job):
        return job.name.split("/")[-1]

    def sync_registered_jobs(self):
        scheduler_backend: BaseSchedulerBackend = app_settings.SCHEDULER_BACKEND_CLASS()
        jobs_list = scheduler_backend.list_jobs()
        unregistered_tasks = set(map(self._get_job_name, jobs_list.jobs)).difference(
            set(self.tasks)
        )

        for task_name in unregistered_tasks:
            if self.schedule_configs.get(task_name):
                del self.schedule_configs[task_name]
            scheduler_backend.delete_job(task_name)

    def get_task(self, name: str) -> Optional[Callable]:
        return self.tasks.get(name)

    def get_schedule_config(self, name: str) -> Optional[Job]:
        return self.schedule_configs.get(name)

    def get_all_tasks(self) -> Dict[str, Callable]:
        return self.tasks


task_registry = TaskRegistry()


def register_task(
    subscriptions: List[str] = [],
    name: Optional[str] = None,
    schedule: Optional[SchedulerJob] = None,
    **kwargs,
) -> Callable[[Callable], Callable]:
    def decorator(f: Callable) -> Callable:
        from .publisher import send_task

        task_name = name or f.__name__

        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return f(*args, **kwargs)

        wrapper.delay = partial(send_task, task_name=task_name)
        wrapper.subscriptions = [subscription.value for subscription in subscriptions]
        wrapper.name = task_name
        if are_subscriptions_valid(subscriptions):
            task_registry.register(wrapper, name=task_name, raw_schedule=schedule)
        setattr(wrapper, TASK_EXTRA_CONTEXT_ATTRIBUTE, kwargs)
        return wrapper

    return decorator



from typing import Any, Callable, Dict, List, Optional

from pydantic import BaseModel, Field


class RequestMessage(BaseModel):
    task_name: str
    args: Optional[List[Any]] = Field(default=[])
    kwargs: Optional[Dict[str, Any]] = Field(default={})


class CallbackContext(BaseModel):
    raw_message: str
    ack: Callable[[], None]
    subscription_name: str


SubscriptionCallback = Callable[[CallbackContext], None]
